/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/

#include <CCore>

#include "RBoard.h"
#include "RFBoard.h"

const long kcoeffStability = 32;
const long kcoeffMobility = 1;

cRFastBoard::StabilityData cRFastBoard::grgStability[256L*256];
cRFastBoard::XCaptureFlag cRFastBoard::grgXCaptureY[8][256];
ulong cRFastBoard::gMinmaxMoveCounterTotal = 0;

#ifdef DUMPANALYSIS

#ifdef MACINTOSH
	#define kszDIRPATHPREFIX ":"
#else
	#define kszDIRPATHPREFIX "\\"
#endif
const char  *cRFastBoard::kszDefaultDumpFile = kszDIRPATHPREFIX"Analysis.txt";
bool cRFastBoard::fDumpMoveAnalysis = true;
ofstream cRFastBoard::ofsDump;

ostream& operator<<(ostream& fs, const cRFastBoard& rfb)
{
	int i,j;

	Assert(cRFastBoard::fDumpMoveAnalysis);

	if (fs.good())
	{
		fs << "Move#" << rfb.CountPieces()-3 << endl;

		JUMPIF(!fs.good(), done);
		
		fs << "  12345678" <<endl;
		for(i=0; i<8;i++)
		{
			const static char rgc[]={'.','_','@', '0'};

			fs << char('a'+i) <<' ';
			for (j=0;j<8;j++)
				fs << rgc[(rfb.Board[i]>>((7-j)*cRFastBoard::kPieceBits))&3];

			fs << endl;

			JUMPIF(!fs.good(), done);
		}
	}
done:
	fs.flush();

	return fs;
}

#endif

void cRFastBoard::
Clear()
{
	super::Clear();

	AClearData(this->BoardY);
	AClearData(this->LegalMoves);
	AClearData(this->LegalMovesBottomFiller);
	AClearData(this->LegalMovesTopFiller);

#define LclrDXY(n) this->BoardXY[n] = (kAllEdges >> (((n)+1)*kPieceBits)) &0xffff
#define RclrDXY(n) this->BoardXY[n] = (kAllEdges << ((15-(n))*kPieceBits)) &0xffff
	LclrDXY(0);LclrDXY(1);LclrDXY(2);LclrDXY(3);LclrDXY(4);LclrDXY(5);LclrDXY(6);LclrDXY(7);
	RclrDXY(8);RclrDXY(9);RclrDXY(10);RclrDXY(11);RclrDXY(12);RclrDXY(13);RclrDXY(14);
#undef RclrDXY
#undef LclrDXY

#define RclrDYX(n) this->BoardYX[n] = (kAllEdges << (((n)+1)*kPieceBits)) &0xffff
#define LclrDYX(n) this->BoardYX[n] = (kAllEdges >> ((15-(n))*kPieceBits)) &0xffff
	RclrDYX(0);RclrDYX(1);RclrDYX(2);RclrDYX(3);RclrDYX(4);RclrDYX(5);RclrDYX(6);RclrDYX(7);
	LclrDYX(8);LclrDYX(9);LclrDYX(10);LclrDYX(11);LclrDYX(12);LclrDYX(13);LclrDYX(14);
#undef LclrDYX
#undef RclrDYX
}

cRFastBoard::
cRFastBoard(const cReversiBoard& rb)
{
	Position i;

	this->Clear();

	memcpy(this->Board, rb.Board,sizeof(this->Board));

	for (i=0; i<8; i++)
		this->BoardY[i] = rb.ExtractY(i);

	for (i=0; i<15; i++)
	{
		this->BoardXY[i] = rb.ExtractXY(i);
		this->BoardYX[i] = rb.ExtractYX(i);
	}
}




inline void cRFastBoard::
StuffColumnFlag(CaptureFlag f, Position column)
{
	Assert(sizeof(this->LegalMoves) == sizeof(uint64));
	*(uint64*)&this->LegalMoves[0] |= *(uint64*)&cRFastBoard::grgXCaptureY[column][f];
}

inline void cRFastBoard::
StuffXYFlag(CaptureFlag rFlag, Position xy)
{
#define DSTFXY(i) this->LegalMoves[xy-i] |= rFlag & ABuildOverlayMask(1, 7-i)
	DSTFXY(7);DSTFXY(6);DSTFXY(5);DSTFXY(4);DSTFXY(3);DSTFXY(2);DSTFXY(1);DSTFXY(0);
#undef DSTFXY
}

inline void cRFastBoard::
StuffYXFlag(CaptureFlag rFlag, Position yx)
{
#define DSTFYX(i) this->LegalMoves[yx+i-7] |= rFlag & ABuildOverlayMask(1, 7-i);
	DSTFYX(0);DSTFYX(1);DSTFYX(2);DSTFYX(3);DSTFYX(4);DSTFYX(5);DSTFYX(6);DSTFYX(7);
#undef DSTFYX
}

void cRFastBoard::
UpdateMobility(Piece p)
{
	const ulong rlPieceNormalizerMask = cRFastBoard::grgPieceNormalizerMask[p];
	const CaptureFlag * const rgCapture = cReversiBoard::grgCaptureFlags;

#define SX(i) this->LegalMoves[i] = rgCapture[this->Board[i] ^ rlPieceNormalizerMask]
		SX(0);SX(1);SX(2);SX(3);SX(4);SX(5);SX(6);SX(7);
#undef SX

#define SMOB(rg, stuff, i)\
		if (rgCapture[this->##rg[i] ^ rlPieceNormalizerMask])\
			this->##stuff(rgCapture[this->##rg[i] ^ rlPieceNormalizerMask], i)

#define SY(i) SMOB(BoardY, StuffColumnFlag, i)
		SY(0);SY(1);SY(2);SY(3);SY(4);SY(5);SY(6);SY(7);
#undef SY

#define SXY(i) SMOB(BoardXY, StuffXYFlag, i)
#define SYX(i) SMOB(BoardYX, StuffYXFlag, i)
		SXY(2);SXY(3);SXY(4);SXY(5);SXY(6);SXY(7);SXY(8);SXY(9);SXY(10);SXY(11);SXY(12);
		SYX(2);SYX(3);SYX(4);SYX(5);SYX(6);SYX(7);SYX(8);SYX(9);SYX(10);SYX(11);SYX(12);
#undef SXY
#undef SYX
#undef SMOB
}
bool cRFastBoard::
NextValidMove(Piece p, Position *px, Position *py) const 
{
	Position x=*px;
	Position y=*py;

	y += (++x)/8;
	x = x%8;
	for(;y<8; y++)
	{
		CaptureFlag rgCapture;

		rgCapture = this->LegalMoves[y];

		for(rgCapture <<= x; rgCapture; x++, rgCapture <<= 1)
		{
			if (rgCapture&0x80)
			{
				*px = x;
				*py = y;
				return true;
			}
		}
		x = 0;
	}

	return false;
}
inline long cRFastBoard::
CountMobility() const 
{
	const ulong *rgBitCount = cRFastBoard::grgBitCount;

#define COUNTMOBILITY(i) rgBitCount[this->LegalMoves[i]]
	
	return 
	COUNTMOBILITY(0)+
	COUNTMOBILITY(1)+
	COUNTMOBILITY(2)+
	COUNTMOBILITY(3)+
	COUNTMOBILITY(4)+
	COUNTMOBILITY(5)+
	COUNTMOBILITY(6)+
	COUNTMOBILITY(7);

#undef COUNTMOBILITY
}

inline long cRFastBoard::
Stability(Piece p) const 
{
	const StabilityData *rgStablity = cRFastBoard::grgStability;

#define SSTB(i) rgStablity[this->Board[i]]
#define SCOL(i) rgStablity[this->BoardY[i]]
#define SXY(i) rgStablity[this->BoardXY[i]]
#define SYX(i) rgStablity[this->BoardYX[i-7]]
		
	long stability = SSTB(0)+SSTB(7)+SCOL(0)+SCOL(7);
/*
		SSTB(0)+SSTB(1)+SSTB(2)+SSTB(3)+SSTB(4)+SSTB(5)+SSTB(6)+SSTB(7)+
		SCOL(0)+SCOL(1)+SCOL(2)+SCOL(3)+SCOL(4)+SCOL(5)+SCOL(6)+SCOL(7)+
		SXY(0)+SXY(1)+SXY(2)+SXY(3)+SXY(4)+SXY(5)+SXY(6)+SXY(7)+SXY(8)+SXY(9)+SXY(10)+SXY(11)+SXY(12)+SXY(13)+SXY(14)+
		SYX(0)+SYX(1)+SYX(2)+SYX(3)+SYX(4)+SYX(5)+SYX(6)+SYX(7)+SYX(9)+SYX(9)+SYX(10)+SYX(11)+SYX(12)+SYX(13)+SYX(14);
*/
	return ((kPiece1-p) * 2 * stability) + stability;//(kPiece1==p)?stability:-stability;

#undef SYX
#undef SXY
#undef SCOL
#undef SSTB
}

long cRFastBoard::
Strength(Piece p) const 
{
//	return kPlusInfinity;
//	return AStandardRand();
//	return this->CountPieces(p)+4*this->CountMobility();
	return 
#ifndef DEBUG
	AStandardRand(-kcoeffMobility, kcoeffMobility) + 
#endif
		this->CountMobility() *kcoeffMobility + this->Stability(p)*kcoeffStability;
//	return this->CountPieces(p);
}

long cRFastBoard::
GreedyStrength(Piece p) const 
{
	return (kInfinity/64) * (this->CountPieces(p) - this->CountPieces(cRFastBoard::OtherPiece(p)));
}

long cRFastBoard::
LosingStrength(Piece p) const 
{
	return -this->Strength(p);
}

long cRFastBoard::
PositionalStrength(Piece p) const 
{
	static const long PA = 6000000;
	static const long PB = -60000;
	static const long PC = 300;
	static const long PD = 10;
	static const long PE = -90000;
	static const long PF = -1000;
	static const long PG = 5;
	static const long PH = 10000;
	static const long PI = 5000;
	static const long PJ = 2500;

	static const long rgWeight[8][8] =
	{
		 {PA,PB,PC,PD,PD,PC,PB,PA},
		 {PB,PE,PF,PG,PG,PF,PE,PB},
		 {PC,PF,PH,PI,PI,PH,PF,PC},
		 {PD,PG,PI,PJ,PJ,PI,PG,PD},
		 {PD,PG,PI,PJ,PJ,PI,PG,PD},
		 {PC,PF,PH,PI,PI,PH,PF,PC},
		 {PB,PE,PF,PG,PG,PF,PE,PB},
		 {PA,PB,PC,PD,PD,PC,PB,PA}
	 };
	const ulong rlPieceNormalizerMask = cRFastBoard::grgPieceNormalizerMask[p];
	long value = 0;

	Position x,y;

	for (y=0; y<8;y++)
	{
		LongLine rLine = this->Board[y]^rlPieceNormalizerMask;

		for(x=0;x<8;x++)
		{
			const Piece p = Piece(rLine&03);

			if(kPiece1 == p)
				value += rgWeight[y][x];
			else if (kPiece2 == p)
				value -= rgWeight[y][x];

			rLine >>= kPieceBits;
		}
	}

	return value;
}

long cRFastBoard::
ABPrune(Piece p, long depth, long useThresh, long passThresh, const bool *pfTerminate)
{
	long movevalue;

	cRFastBoard::gMinmaxMoveCounterTotal++;

	this->UpdateMobility(p);
	if ((depth--<=0) || (*pfTerminate))
	{
		//movevalue = this->BoardStrength(p);//gReversiMinmaxMoveCounter;
		movevalue = (this->*pStrengthFunc)(p);
	}
	else
	{
		Position x, y;
		long newvalue;
		Piece otherpiece = cRFastBoard::OtherPiece(p);

		x = -1;
		y = 0;
		if (!this->NextValidMove(p, &x,&y))
		{
			x=-1;
			y=0;

			this->UpdateMobility(otherpiece);
			if (!this->NextValidMove(otherpiece, &x,&y))
			{
				movevalue = this->GreedyStrength(p);
			}
			else
			{
				do
				{
					cRFastBoard tBoard= *this;

					tBoard.MakeMove(otherpiece, x, y);
					newvalue = tBoard.ABPrune(p, depth, useThresh, passThresh, pfTerminate);

					if (newvalue < useThresh)
						useThresh = newvalue;

					if (useThresh <= passThresh)
						break;
				}
				while(this->NextValidMove(otherpiece, &x,&y));
				movevalue = useThresh;
			}
		}
		else
		{
			do
			{
				cRFastBoard tBoard= *this;

				tBoard.MakeMove(p, x, y);
				newvalue = -tBoard.ABPrune(otherpiece, depth, -passThresh, -useThresh, pfTerminate);
				if (newvalue > passThresh)
					passThresh = newvalue;

				if (passThresh >= useThresh)
					break;
			}
			while(this->NextValidMove(p, &x,&y));
			movevalue = passThresh;
		}
	}
#ifdef DEBUGELEPHANT
//	ATraceLn("L%d, P%d, pass:%d, use:%d", depth, (int)p, passThresh, useThresh);

	if (cRFastBoard::fDumpMoveAnalysis)
	{
		int i;
		
		for (i=0; i<(5-depth); i++)
			cRFastBoard::ofsDump << "......";
		cRFastBoard::ofsDump<< movevalue<<endl;
	}
#endif

	ThreadYieldAfterPeriod(AGetThreadTimeoutFrequency());
	
	return movevalue;
}

